@@ -7,7 +7,7 @@
 #
 # Author: Michael Rash (mbr@cipherdyne.org)
 #
-# Version: 1.3.1
+# Version: 1.6.1
 #
 ##################################################################
 #
@@ -17,27 +17,35 @@ package IPTables::Parse;
 use 5.006;
 use POSIX ":sys_wait_h";
 use Carp;
+use File::Temp;
 use strict;
 use warnings;
 use vars qw($VERSION);
 
-$VERSION = '1.3.1';
+$VERSION = '1.6.1';
 
 sub new() {
     my $class = shift;
     my %args  = @_;
 
-    my $ipt_bin  = '/sbin/iptables';
-    my $ipt6_bin = '/sbin/ip6tables';
-    my $fwc_bin  = '/usr/bin/firewall-cmd';
+    ### default iptables/ip6tables/firewall-cmd paths
+    my $ipt_bin    = '/sbin/iptables';
+    my $ipt6_bin   = '/sbin/ip6tables';
+    my $fwc_bin    = '/usr/bin/firewall-cmd';
+
+    my $iptout_pat = 'ipt.out.XXXXXX';
+    my $ipterr_pat = 'ipt.err.XXXXXX';
 
     my $self = {
-        _iptables        => $args{'iptables'} || $args{'ip6tables'} || '',
+        _iptables        => $args{'iptables'}     || $args{'ip6tables'} || '',
         _firewall_cmd    => $args{'firewall-cmd'} || '',
         _fwd_args        => $args{'fwd_args'}     || '--direct --passthrough ipv4',
         _ipv6            => $args{'use_ipv6'}     || 0,
-        _iptout          => $args{'iptout'}       || '/tmp/ipt.out',
-        _ipterr          => $args{'ipterr'}       || '/tmp/ipt.err',
+        _iptout          => $args{'iptout'}       || '',
+        _iptout_pat      => $args{'iptout_pat'}   || '',
+        _ipterr          => $args{'ipterr'}       || '',
+        _ipterr_pat      => $args{'ipterr_pat'}   || '',
+        _tmpdir          => $args{'tmpdir'}       || '',
         _ipt_alarm       => $args{'ipt_alarm'}    || 30,
         _debug           => $args{'debug'}        || 0,
         _verbose         => $args{'verbose'}      || 0,
@@ -45,10 +53,11 @@ sub new() {
         _ipt_exec_style  => $args{'ipt_exec_style'}  || 'waitpid',
         _ipt_exec_sleep  => $args{'ipt_exec_sleep'}  || 0,
         _sigchld_handler => $args{'sigchld_handler'} || \&REAPER,
-        _skip_ipt_exec_check => $args{'skip_ipt_exec_check'} || 0
+        _skip_ipt_exec_check => $args{'skip_ipt_exec_check'} || 0,
+        _lockless_ipt_exec   => $args{'lockless_ipt_exec'}   || 0,
     };
 
-    if ($self->{'_skip_ipt_exec_check'}) {
+    if ($self->{'_skip_ipt_exec_check'}) {  ### useful for file-only parsing
         unless ($self->{'_firewall_cmd'} or $self->{'_iptables'}) {
             ### default
             $self->{'_iptables'} = $ipt_bin;
@@ -68,28 +77,70 @@ sub new() {
             ### check for firewall-cmd first since systems with it
             ### will have iptables installed as well (but firewall-cmd
             ### should be used instead if it exists)
-            if (-e $fwc_bin and -x $fwc_bin) {
-                $self->{'_firewall_cmd'} = $fwc_bin;
-            } elsif (-e $ipt_bin and -x $ipt_bin) {
-                $self->{'_iptables'} = $ipt_bin;
-            } elsif (-e $ipt6_bin and -x $ipt6_bin) {
-                $self->{'_iptables'} = $ipt6_bin;
-            } else {
+            my $found = 0;
+            for my $cmd ('firewall-cmd', 'iptables', 'ip6tables') {
+                next if $self->{'_ipv6'} and $cmd eq 'iptables';
+                my $path = &check_cmd($cmd);
+                if ($path) {
+                    if ($cmd eq 'firewall-cmd') {
+                        $self->{'_firewall_cmd'} = $path;
+                    } else {
+                        $self->{'_iptables'} = $path;
+                    }
+                    $found = 1;
+                    last;
+                }
+            }
+            unless ($found) {
                 croak "[*] Could not find/execute iptables, " .
-                    "specify path via _iptables\n";
+                    "specify path via 'iptables' key.\n";
             }
         }
     }
 
-    if ($self->{'_ipv6'} and $self->{'_iptables'} eq $ipt_bin) {
+    if ($self->{'_ipv6'} and $self->{'_iptables'} !~ /ip6tables/) {
         if (-e $ipt6_bin and -x $ipt6_bin) {
             $self->{'_iptables'} = $ipt6_bin;
         } else {
-            croak "[*] Could not find/execute ip6tables, " .
-                "specify path via _iptables\n";
+            my $path = &check_cmd('ip6tables');
+            if ($path) {
+                $self->{'_iptables'} = $path;
+            } else {
+                croak "[*] Could not find/execute ip6tables, " .
+                    "specify path via 'ip6tables' hash key.\n";
+            }
         }
     }
 
+    ### set up the path for temporary files
+    if ($self->{'_tmpdir'} and -d $self->{'_tmpdir'}) {
+        if ($self->{'_iptout_pat'}) {
+            $self->{'_iptout'}
+                = mktemp("$self->{'_tmpdir'}/$self->{'_iptout_pat'}");
+        } elsif ($self->{'_iptout'}) {
+            $self->{'_iptout'} = mktemp("$self->{'_tmpdir'}/$self->{'_iptout'}");
+        } else {
+            $self->{'_iptout'} = mktemp("$self->{'_tmpdir'}/$iptout_pat");
+        }
+        if ($self->{'_ipterr_pat'}) {
+            $self->{'_ipterr'}
+                = mktemp("$self->{'_tmpdir'}/$self->{'_iptout_pat'}");
+        } elsif ($self->{'_ipterr'}) {
+            $self->{'_ipterr'} = mktemp("$self->{'_tmpdir'}/$self->{'_ipterr'}");
+        } else {
+            $self->{'_ipterr'} = mktemp("$self->{'_tmpdir'}/$ipterr_pat");
+        }
+    } else {
+        croak "[*] 'iptout_pat' is only valid with 'tmpdir' set."
+            if $self->{'_iptout_pat'};
+        croak "[*] 'ipterr_pat' is only valid with 'tmpdir' set."
+            if $self->{'_iptout_err'};
+        $self->{'_iptout'} = mktemp("/tmp/$iptout_pat")
+            unless $self->{'_iptout'};
+        $self->{'_ipterr'} = mktemp("/tmp/$ipterr_pat")
+            unless $self->{'_ipterr'};
+    }
+
     ### set the firewall binary name
     $self->{'_ipt_bin_name'} = 'iptables';
     if ($self->{'_firewall_cmd'}) {
@@ -108,7 +159,7 @@ sub new() {
             if ($self->{'_ipt_bin_name'} eq 'iptables') {
                 unless ($self->{'_skip_ipt_exec_check'}) {
                     croak "[*] use_ipv6 is true, " .
-                        "but $self->{'_iptables'} not ip6tables.\n";
+                        "but $self->{'_iptables'} is not ip6tables.\n";
                 }
             }
         }
@@ -126,11 +177,33 @@ sub new() {
         $self->{'_cmd'} = "$self->{'_firewall_cmd'} $self->{'_fwd_args'}";
     }
 
+    unless ($self->{'_skip_ipt_exec_check'}) {
+        unless ($self->{'_lockless_ipt_exec'}) {
+            ### now that we have the iptables command defined, see whether
+            ### it supports -w to acquire an exclusive lock
+            my ($rv, $out_ar, $err_ar) = &exec_iptables($self,
+                "$self->{'_cmd'} -w -t filter -n -L INPUT");
+            $self->{'_cmd'} .= ' -w' if $rv;
+        }
+    }
+
     $self->{'parse_keys'} = &parse_keys();
 
     bless $self, $class;
 }
 
+sub DESTROY {
+    my $self = shift;
+
+    ### clean up tmp files
+    unless ($self->{'_debug'}) {
+        unlink $self->{'_iptout'};
+        unlink $self->{'_ipterr'};
+    }
+
+    return;
+}
+
 sub parse_keys() {
     my $self = shift;
 
@@ -228,12 +301,29 @@ sub parse_keys() {
                 'ipt_match'  => '-m length --length',
             },
         },
+        'rule_num' => '',
         'raw' => ''
     );
 
     return \%keys;
 }
 
+sub check_cmd() {
+    my $cmd_name = shift;
+
+    my $path = '';
+
+    for my $search_path ('/sbin', '/usr/sbin', '/bin', '/usr/bin') {
+        my $test_path = "$search_path/$cmd_name";
+        if (-e $test_path and -x $test_path) {
+            $path = $test_path;
+            last;
+        }
+    }
+
+    return $path;
+}
+
 sub list_table_chains() {
     my $self   = shift;
     my $table  = shift || croak '[*] Specify a table, e.g. "nat"';
@@ -259,7 +349,7 @@ sub list_table_chains() {
     }
 
     for (@ipt_lines) {
-        if (/^\s*Chain\s+(\w+)/) {
+        if (/^\s*Chain\s(.*?)\s\(/) {
             push @chains, $1;
         }
     }
@@ -316,6 +406,9 @@ sub chain_rules() {
     my $found_chain  = 0;
     my @ipt_lines = ();
 
+    my $fh = *STDERR;
+    $fh = *STDOUT if $self->{'_verbose'};
+
     ### only used for IPv4 + NAT
     my $ip_re = qr|(?:[0-2]?\d{1,2}\.){3}[0-2]?\d{1,2}|;
 
@@ -335,7 +428,7 @@ sub chain_rules() {
         close F;
     } else {
         my ($rv, $out_ar, $err_ar) = $self->exec_iptables(
-                "$self->{'_cmd'} -t $table -v -n -L $chain");
+                "$self->{'_cmd'} -t $table -v -n -L $chain --line-numbers");
         @ipt_lines = @$out_ar;
     }
 
@@ -343,33 +436,43 @@ sub chain_rules() {
     ### policy data came from a file then -v might not have been used)
     my $ipt_verbose = 0;
     for my $line (@ipt_lines) {
-        if ($line =~ /^\s*pkts\s+bytes\s+target/) {
+        if ($line =~ /\spkts\s+bytes\s+target/) {
             $ipt_verbose = 1;
             last;
         }
     }
+    my $has_line_numbers = 0;
+    for my $line (@ipt_lines) {
+        if ($line =~ /^num\s+pkts\s+bytes\s+target/) {
+            $has_line_numbers = 1;
+            last;
+        }
+    }
+
+    my $rule_num = 0;
 
     LINE: for my $line (@ipt_lines) {
         chomp $line;
 
         last LINE if ($found_chain and $line =~ /^\s*Chain\s+/);
 
-        if ($line =~ /^\s*Chain\s+$chain\s+\(/i) {
+        if ($line =~ /^\s*Chain\s\Q$chain\E\s\(/i) {
             $found_chain = 1;
             next LINE;
         }
-        if ($ipt_verbose) {
-            next LINE if $line =~ /^\s*pkts\s+bytes\s+target\s/i;
-        } else {
-            next LINE if $line =~ /^\s*target\s+prot/i;
-        }
+        next LINE if $line =~ /\starget\s{2,}prot/i;
         next LINE unless $found_chain;
         next LINE unless $line;
 
+        ### track the rule number independently of --line-numbers,
+        ### but the values should always match
+        $rule_num++;
+
         ### initialize hash
         my %rule = (
             'extended' => '',
-            'raw'      => $line
+            'raw'      => $line,
+            'rule_num' => $rule_num
         );
         for my $key (keys %{$self->{'parse_keys'}->{'regular'}}) {
             $rule{$key} = '';
@@ -378,6 +481,43 @@ sub chain_rules() {
             $rule{$key} = '';
         }
 
+        my $rule_body = '';
+        my $packets   = '';
+        my $bytes     = '';
+        my $rnum      = '';
+
+        if ($ipt_verbose) {
+            if ($has_line_numbers) {
+                if ($line =~ /^\s*(\d+)\s+(\S+)\s+(\S+)\s+(.*)/) {
+                    $rnum      = $1;
+                    $packets   = $2;
+                    $bytes     = $3;
+                    $rule_body = $4;
+                }
+            } else {
+                if ($line =~ /^\s*(\S+)\s+(\S+)\s+(.*)/) {
+                    $packets   = $1;
+                    $bytes     = $2;
+                    $rule_body = $3;
+                }
+            }
+        } else {
+            if ($has_line_numbers) {
+                if ($line =~ /^\s*(\d+)\s+(.*)/) {
+                    $rnum      = $1;
+                    $rule_body = $2;
+                }
+            } else {
+                $rule_body = $line;
+                $rnum      = $rule_num;
+                $rnum      = $rule_num;
+            }
+        }
+
+        if ($rnum and $rnum ne $rule_num) {
+            croak "[*] Rule number mis-match.";
+        }
+
         if ($ipt_verbose) {
 
             ### iptables:
@@ -391,28 +531,34 @@ sub chain_rules() {
             ### 0     0 ACCEPT  tcp   *   *   ::/0     fe80::aa:0:1/128    tcp dpt:12345
             ### 0     0 LOG     all   *   *   ::/0     ::/0                LOG flags 0 level 4
 
-            my $match_re = qr/^\s*(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+\-\-\s+
+            my $match_re = qr/^(\S+)\s+(\S+)\s+\-\-\s+
                                 (\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*(.*)/x;
 
             if ($self->{'_ipt_bin_name'} eq 'ip6tables'
                     or ($self->{'_ipt_bin_name'} eq 'firewall-cmd'
                     and $self->{'_fwd_args'} =~ /\sipv6/)) {
-                $match_re = qr/^\s*(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+
+                $match_re = qr/^(\S+)\s+(\S+)\s+
                                 (\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*(.*)/x;
             }
 
-            if ($line =~ $match_re) {
-                $rule{'packets'}  = $1;
-                $rule{'bytes'}    = $2;
-                $rule{'target'}   = $3;
-                $rule{'protocol'} = $rule{'proto'} = lc($4);
-                $rule{'intf_in'}  = $5;
-                $rule{'intf_out'} = $6;
-                $rule{'src'}      = $7;
-                $rule{'dst'}      = $8;
-                $rule{'extended'} = $9 || '';
+            if ($rule_body =~ $match_re) {
+                $rule{'packets'}  = $packets;
+                $rule{'bytes'}    = $bytes;
+                $rule{'target'}   = $1;
+                my $proto = $2;
+                $proto = 'all' if $proto eq '0';
+                $rule{'protocol'} = $rule{'proto'} = lc($proto);
+                $rule{'intf_in'}  = $3;
+                $rule{'intf_out'} = $4;
+                $rule{'src'}      = $5;
+                $rule{'dst'}      = $6;
+                $rule{'extended'} = $7 || '';
 
                 &parse_rule_extended(\%rule, $self->{'parse_keys'}->{'extended'});
+            } else {
+                if ($self->{'_debug'}) {
+                    print $fh localtime() . "     -v Did not match parse regex: $line\n";
+                }
             }
         } else {
 
@@ -433,24 +579,28 @@ sub chain_rules() {
             ### ACCEPT     tcp   ::/0     fe80::aa:0:1/128    tcp dpt:12345
             ### LOG        all   ::/0     ::/0                LOG flags 0 level 4
 
-            my $match_re = qr/^\s*(\S+)\s+(\S+)\s+\-\-\s+(\S+)\s+(\S+)\s*(.*)/;
+            my $match_re = qr/^(\S+)\s+(\S+)\s+\-\-\s+(\S+)\s+(\S+)\s*(.*)/;
 
             if ($self->{'_ipt_bin_name'} eq 'ip6tables'
                     or ($self->{'_ipt_bin_name'} eq 'firewall-cmd'
                     and $self->{'_fwd_args'} =~ /\sipv6/)) {
-                $match_re = qr/^\s*(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*(.*)/;
+                $match_re = qr/^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*(.*)/;
             }
 
-            if ($line =~ $match_re) {
+            if ($rule_body =~ $match_re) {
                 $rule{'target'}   = $1;
                 my $proto = $2;
                 $proto = 'all' if $proto eq '0';
-                $rule{'protocol'} = $rule{'proto'} = $proto;
+                $rule{'protocol'} = $rule{'proto'} = lc($proto);
                 $rule{'src'}      = $3;
                 $rule{'dst'}      = $4;
                 $rule{'extended'} = $5 || '';
 
                 &parse_rule_extended(\%rule, $self->{'parse_keys'}->{'extended'});
+            } else {
+                if ($self->{'_debug'}) {
+                    print $fh localtime() . "     Did not match parse regex: $line\n";
+                }
             }
         }
         push @chain, \%rule;
@@ -722,7 +872,7 @@ sub sub_chains() {
         chomp $line;
         ### Chain INPUT (policy DROP)
         ### Chain fwsnort_INPUT_eth1 (1 references)
-        if ($line =~ /^\s*Chain\s+$start_chain\s+\(/ and
+        if ($line =~ /^\s*Chain\s+\Q$start_chain\E\s+\(/ and
                 $line !~ /0\s+references/) {
             $found = 1;
             next;
@@ -771,6 +921,12 @@ sub exec_iptables() {
             or $cmd =~ m|^\s*ip6tables| or $cmd =~ m|^\S+/ip6tables|
             or $cmd =~ m|^\s*firewall-cmd| or $cmd =~ m|^\S+/firewall-cmd|;
 
+    ### sanitize $cmd - this is not bullet proof, but better than
+    ### nothing (especially for strange iptables chain names). Further,
+    ### quotemeta() is too aggressive since things like IPv6 addresses
+    ### contain ":" chars, etc.
+    $cmd =~ s/([;<>\$\|`\@&\(\)\[\]\{\}])/\\$1/g;
+
     my $rv = 1;
     my @stdout = ();
     my @stderr = ();
@@ -904,6 +1060,8 @@ sub REAPER {
 1;
 __END__
 
+=encoding UTF-8
+
 =head1 NAME
 
 IPTables::Parse - Perl extension for parsing iptables and ip6tables policies
@@ -916,8 +1074,6 @@ IPTables::Parse - Perl extension for parsing iptables and ip6tables policies
       'use_ipv6' => 0,         # can set to 1 to force ip6tables usage
       'ipt_rules_file' => '',  # optional file path from
                                # which to read iptables rules
-      'iptout'   => '/tmp/iptables.out',
-      'ipterr'   => '/tmp/iptables.err',
       'debug'    => 0,
       'verbose'  => 0
   );
@@ -982,7 +1138,8 @@ iptables/ip6tables commands, or from parsing a file that contains an
 iptables/ip6tables policy listing. Note that the 'firewalld' infrastructure on
 Fedora21 is also supported through execution of the 'firewall-cmd' binary.
 By default, the path to iptables is assumed to be '/sbin/iptables', but if the
-firewall is 'firewalld', then the '/usr/bin/firewall-cmd' is used.
+firewall is 'firewalld', then the '/usr/bin/firewall-cmd' is used. Both of
+these paths are configurable via the keys mentioned below.
 
 With this module, you can get the current policy applied to a
 table/chain, look for a specific user-defined chain, check for a default DROP
@@ -993,18 +1150,24 @@ Note that if you initialize the IPTables::Parse object with the 'ipt_rules_file'
 key, then all parsing routines will open the specified file for iptables rules
 data. So, you can create this file with a command like
 'iptables -t filter -nL -v > ipt.rules', and then initialize the object with
-IPTables::Parse->new({'ipt_rules_file'=>'ipt.rules'}). Further, if you are
+IPTables::Parse->new('ipt_rules_file' => 'ipt.rules'). Further, if you are
 running on a system without iptables installed, but you have an iptables policy
 written to the ipt.rules file, then you can pass in 'skip_ipt_exec_check=>1'
 in order to analyze the file without having IPTables::Parse check for the
 iptables binary.
 
 In summary, in addition to the hash keys mentioned above, optional keys that
-can be passed to new() include '_iptables' (set path to iptables binary),
-'_firewall_cmd' (set path to 'firewall-cmd' binary for systems with
-'firewalld'), '_fwd_args' (set 'firewall-cmd' usage args; defaults to
-'--direct --passthrough ipv4'), '_ipv6' (set IPv6 mode for ip6tables),
-'_debug' and '_verbose'.
+can be passed to new() include 'iptables' (set path to iptables binary),
+'firewall_cmd' (set path to 'firewall-cmd' binary for systems with
+'firewalld'), 'fwd_args' (set 'firewall-cmd' usage args; defaults to
+'--direct --passthrough ipv4'), 'ipv6' (set IPv6 mode for ip6tables),
+'iptout' (set path to temporary stdout file, defaults to /tmp/ipt.out.XXXXXX),
+'iptout_pat' (set pattern for temporary stdout file in the 'tmpdir' directory),
+'ipterr' (set path to temporary stderr file, defaults to /tmp/ipt.err.XXXXXX),
+'iptout_err' (set pattern for temporary stderr file in the 'tmpdir' directory),
+'tmpdir' (set path to temporary file handling directory),
+'debug', 'verbose', and 'lockless_ipt_exec' (disable usage of the iptables
+'-w' argument that acquires an exclusive lock on command execution).
 
 =head1 FUNCTIONS
 
@@ -1113,8 +1276,10 @@ Source control is provided by git:
 Thanks to the following people:
 
   Franck Joncourt <franck.mail@dthconnex.com>
+  Stuart Schneider
   Grant Ferley
   Fabien Mazieres
+  Miloslav Trmač
 
 =head1 AUTHOR
 
@@ -1124,7 +1289,7 @@ this address if there are any questions, comments, or bug reports.
 
 =head1 VERSION
 
-Version 1.3 (February, 2015)
+Version 1.6.1 (November, 2015)
 
 =head1 COPYRIGHT AND LICENSE
 
